package parser

import "strings"

// Exceptions are blocks of the PSL that would fail current validation
// and stylistic requirements, but are exempted due to predating those
// rules.
//
// These exceptions are deliberately built to be brittle: editing a
// block revokes its exemptions and requires the block to pass all
// modern validations (or the exceptions below need to be
// updated). This hopefully ratchets the PSL to always become more
// conformant with current policy, while not requiring that all
// existing lint be fixed immediately.
//
// See the bottom of this file for the exceptions themselves.

// downgradeToWarning reports whether e is a legacy exception to
// normal parsing and validation rules, and should be reported as a
// warning rather than a validation error.
func downgradeToWarning(e error) bool {
	switch v := e.(type) {
	case MissingEntityEmail:
		return sourceIsExempted(missingEmail, v.Suffixes.Text())
	}
	return false
}

// exemptFromSorting reports whether the block owned by entity is
// exempt from the sorting requirement that normally applies in the
// private domains section.
func exemptFromSorting(source Source) bool {
	return sourceIsExempted(incorrectSort, source.Text())
}

func sourceIsExempted(exceptions []string, source string) bool {
	for _, exc := range exceptions {
		if exc == source {
			return true
		}
	}
	return false
}

func lines(lines ...string) string {
	return strings.Join(lines, "\n")
}

// missingEmail are source code blocks in the private domains section
// that are allowed to lack email contact information.
var missingEmail = []string{
	lines(
		"// 611coin : https://611project.org/",
		"611.to",
	),
	lines(
		"// c.la : http://www.c.la/",
		"c.la",
	),
	lines(
		"// co.ca : http://registry.co.ca/",
		"co.ca",
	),
	lines(
		"// DynDNS.com : http://www.dyndns.com/services/dns/dyndns/",
		"dyndns.biz",
		"for-better.biz",
		"for-more.biz",
		"for-some.biz",
		"for-the.biz",
		"selfip.biz",
		"webhop.biz",
		"ftpaccess.cc",
		"game-server.cc",
		"myphotos.cc",
		"scrapping.cc",
		"blogdns.com",
		"cechire.com",
		"dnsalias.com",
		"dnsdojo.com",
		"doesntexist.com",
		"dontexist.com",
		"doomdns.com",
		"dyn-o-saur.com",
		"dynalias.com",
		"dyndns-at-home.com",
		"dyndns-at-work.com",
		"dyndns-blog.com",
		"dyndns-free.com",
		"dyndns-home.com",
		"dyndns-ip.com",
		"dyndns-mail.com",
		"dyndns-office.com",
		"dyndns-pics.com",
		"dyndns-remote.com",
		"dyndns-server.com",
		"dyndns-web.com",
		"dyndns-wiki.com",
		"dyndns-work.com",
		"est-a-la-maison.com",
		"est-a-la-masion.com",
		"est-le-patron.com",
		"est-mon-blogueur.com",
		"from-ak.com",
		"from-al.com",
		"from-ar.com",
		"from-ca.com",
		"from-ct.com",
		"from-dc.com",
		"from-de.com",
		"from-fl.com",
		"from-ga.com",
		"from-hi.com",
		"from-ia.com",
		"from-id.com",
		"from-il.com",
		"from-in.com",
		"from-ks.com",
		"from-ky.com",
		"from-ma.com",
		"from-md.com",
		"from-mi.com",
		"from-mn.com",
		"from-mo.com",
		"from-ms.com",
		"from-mt.com",
		"from-nc.com",
		"from-nd.com",
		"from-ne.com",
		"from-nh.com",
		"from-nj.com",
		"from-nm.com",
		"from-nv.com",
		"from-oh.com",
		"from-ok.com",
		"from-or.com",
		"from-pa.com",
		"from-pr.com",
		"from-ri.com",
		"from-sc.com",
		"from-sd.com",
		"from-tn.com",
		"from-tx.com",
		"from-ut.com",
		"from-va.com",
		"from-vt.com",
		"from-wa.com",
		"from-wi.com",
		"from-wv.com",
		"from-wy.com",
		"getmyip.com",
		"gotdns.com",
		"hobby-site.com",
		"homelinux.com",
		"homeunix.com",
		"iamallama.com",
		"is-a-anarchist.com",
		"is-a-blogger.com",
		"is-a-bookkeeper.com",
		"is-a-bulls-fan.com",
		"is-a-caterer.com",
		"is-a-chef.com",
		"is-a-conservative.com",
		"is-a-cpa.com",
		"is-a-cubicle-slave.com",
		"is-a-democrat.com",
		"is-a-designer.com",
		"is-a-doctor.com",
		"is-a-financialadvisor.com",
		"is-a-geek.com",
		"is-a-green.com",
		"is-a-guru.com",
		"is-a-hard-worker.com",
		"is-a-hunter.com",
		"is-a-landscaper.com",
		"is-a-lawyer.com",
		"is-a-liberal.com",
		"is-a-libertarian.com",
		"is-a-llama.com",
		"is-a-musician.com",
		"is-a-nascarfan.com",
		"is-a-nurse.com",
		"is-a-painter.com",
		"is-a-personaltrainer.com",
		"is-a-photographer.com",
		"is-a-player.com",
		"is-a-republican.com",
		"is-a-rockstar.com",
		"is-a-socialist.com",
		"is-a-student.com",
		"is-a-teacher.com",
		"is-a-techie.com",
		"is-a-therapist.com",
		"is-an-accountant.com",
		"is-an-actor.com",
		"is-an-actress.com",
		"is-an-anarchist.com",
		"is-an-artist.com",
		"is-an-engineer.com",
		"is-an-entertainer.com",
		"is-certified.com",
		"is-gone.com",
		"is-into-anime.com",
		"is-into-cars.com",
		"is-into-cartoons.com",
		"is-into-games.com",
		"is-leet.com",
		"is-not-certified.com",
		"is-slick.com",
		"is-uberleet.com",
		"is-with-theband.com",
		"isa-geek.com",
		"isa-hockeynut.com",
		"issmarterthanyou.com",
		"likes-pie.com",
		"likescandy.com",
		"neat-url.com",
		"saves-the-whales.com",
		"selfip.com",
		"sells-for-less.com",
		"sells-for-u.com",
		"servebbs.com",
		"simple-url.com",
		"space-to-rent.com",
		"teaches-yoga.com",
		"writesthisblog.com",
		"ath.cx",
		"fuettertdasnetz.de",
		"isteingeek.de",
		"istmein.de",
		"lebtimnetz.de",
		"leitungsen.de",
		"traeumtgerade.de",
		"barrel-of-knowledge.info",
		"barrell-of-knowledge.info",
		"dyndns.info",
		"for-our.info",
		"groks-the.info",
		"groks-this.info",
		"here-for-more.info",
		"knowsitall.info",
		"selfip.info",
		"webhop.info",
		"forgot.her.name",
		"forgot.his.name",
		"at-band-camp.net",
		"blogdns.net",
		"broke-it.net",
		"buyshouses.net",
		"dnsalias.net",
		"dnsdojo.net",
		"does-it.net",
		"dontexist.net",
		"dynalias.net",
		"dynathome.net",
		"endofinternet.net",
		"from-az.net",
		"from-co.net",
		"from-la.net",
		"from-ny.net",
		"gets-it.net",
		"ham-radio-op.net",
		"homeftp.net",
		"homeip.net",
		"homelinux.net",
		"homeunix.net",
		"in-the-band.net",
		"is-a-chef.net",
		"is-a-geek.net",
		"isa-geek.net",
		"kicks-ass.net",
		"office-on-the.net",
		"podzone.net",
		"scrapper-site.net",
		"selfip.net",
		"sells-it.net",
		"servebbs.net",
		"serveftp.net",
		"thruhere.net",
		"webhop.net",
		"merseine.nu",
		"mine.nu",
		"shacknet.nu",
		"blogdns.org",
		"blogsite.org",
		"boldlygoingnowhere.org",
		"dnsalias.org",
		"dnsdojo.org",
		"doesntexist.org",
		"dontexist.org",
		"doomdns.org",
		"dvrdns.org",
		"dynalias.org",
		"dyndns.org",
		"go.dyndns.org",
		"home.dyndns.org",
		"endofinternet.org",
		"endoftheinternet.org",
		"from-me.org",
		"game-host.org",
		"gotdns.org",
		"hobby-site.org",
		"homedns.org",
		"homeftp.org",
		"homelinux.org",
		"homeunix.org",
		"is-a-bruinsfan.org",
		"is-a-candidate.org",
		"is-a-celticsfan.org",
		"is-a-chef.org",
		"is-a-geek.org",
		"is-a-knight.org",
		"is-a-linux-user.org",
		"is-a-patsfan.org",
		"is-a-soxfan.org",
		"is-found.org",
		"is-lost.org",
		"is-saved.org",
		"is-very-bad.org",
		"is-very-evil.org",
		"is-very-good.org",
		"is-very-nice.org",
		"is-very-sweet.org",
		"isa-geek.org",
		"kicks-ass.org",
		"misconfused.org",
		"podzone.org",
		"readmyblog.org",
		"selfip.org",
		"sellsyourhome.org",
		"servebbs.org",
		"serveftp.org",
		"servegame.org",
		"stuff-4-sale.org",
		"webhop.org",
		"better-than.tv",
		"dyndns.tv",
		"on-the-web.tv",
		"worse-than.tv",
		"is-by.us",
		"land-4-sale.us",
		"stuff-4-sale.us",
		"dyndns.ws",
		"mypets.ws",
	),
	lines(
		"// Hashbang : https://hashbang.sh",
		"hashbang.sh",
	),
	lines(
		"// HostyHosting (https://hostyhosting.com)",
		"hostyhosting.io",
	),
	lines(
		"// info.at : http://www.info.at/",
		"biz.at",
		"info.at",
	),
	lines(
		"// .KRD : http://nic.krd/data/krd/Registration%20Policy.pdf",
		"co.krd",
		"edu.krd",
	),
	lines(
		"// Michau Enterprises Limited : http://www.co.pl/",
		"co.pl",
	),
	lines(
		"// Nicolaus Copernicus University in Torun - MSK TORMAN (https://www.man.torun.pl)",
		"torun.pl",
	),
	lines(
		"// TASK geographical domains (https://www.task.gda.pl/uslugi/dns)",
		"gda.pl",
		"gdansk.pl",
		"gdynia.pl",
		"med.pl",
		"sopot.pl",
	),
	lines(
		"// CoDNS B.V.",
		"co.nl",
		"co.no",
	),
	lines(
		"// .pl domains (grandfathered)",
		"art.pl",
		"gliwice.pl",
		"krakow.pl",
		"poznan.pl",
		"wroc.pl",
		"zakopane.pl",
	),
	lines(
		"// QA2",
		"// Submitted by Daniel Dent (https://www.danieldent.com/)",
		"qa2.com",
	),
}

// incorrectSort are source code blocks in the private domains section
// that are allowed to be in the wrong sort order.
var incorrectSort = []string{
	lines(
		"// AAA workspace : https://aaa.vodka",
		"// Submitted by Kirill Rezraf <admin@aaa.vodka>",
		"aaa.vodka",
	),
	lines(
		"// University of Banja Luka : https://unibl.org",
		"// Domains for Republic of Srpska administrative entity.",
		"// Submitted by Marko Ivanovic <kormang@hotmail.rs>",
		"rs.ba",
	),
	lines(
		"// University of Bielsko-Biala regional domain: http://dns.bielsko.pl/",
		"// Submitted by Marcin <dns@ath.bielsko.pl>",
		"bielsko.pl",
	),
	lines(
		"// No longer operated by CentralNic, these entries should be adopted and/or removed by current operators",
		"// Submitted by Gavin Brown <gavin.brown@centralnic.com>",
		"ar.com",
		"hu.com",
		"kr.com",
		"no.com",
		"qc.com",
		"uy.com",
	),
	lines(
		"// Africa.com Web Solutions Ltd : https://registry.africa.com",
		"// Submitted by Gavin Brown <gavin.brown@centralnic.com>",
		"africa.com",
	),
	lines(
		"// iDOT Services Limited : http://www.domain.gr.com",
		"// Submitted by Gavin Brown <gavin.brown@centralnic.com>",
		"gr.com",
	),
	lines(
		"// Radix FZC : http://domains.in.net",
		"// Submitted by Gavin Brown <gavin.brown@centralnic.com>",
		"web.in",
		"in.net",
	),
	lines(
		"// US REGISTRY LLC : http://us.org",
		"// Submitted by Gavin Brown <gavin.brown@centralnic.com>",
		"us.org",
	),
	lines(
		"// co.com Registry, LLC : https://registry.co.com",
		"// Submitted by Gavin Brown <gavin.brown@centralnic.com>",
		"co.com",
	),
	lines(
		"// Roar Domains LLC : https://roar.basketball/",
		"// Submitted by Gavin Brown <gavin.brown@centralnic.com>",
		"aus.basketball",
		"nz.basketball",
	),
	lines(
		"// BRS Media : https://brsmedia.com/",
		"// Submitted by Gavin Brown <gavin.brown@centralnic.com>",
		"radio.am",
		"radio.fm",
	),
	lines(
		"// c.la : http://www.c.la/",
		"c.la",
	),
	lines(
		"// Clever Cloud : https://www.clever-cloud.com/",
		"// Submitted by Quentin Adam <noc@clever-cloud.com>",
		"cleverapps.cc",
		"*.services.clever-cloud.com",
		"cleverapps.io",
		"cleverapps.tech",
	),
	lines(
		"// co.ca : http://registry.co.ca/",
		"co.ca",
	),
	lines(
		"// Co & Co : https://co-co.nl/",
		"// Submitted by Govert Versluis <govert@co-co.nl>",
		"*.otap.co",
	),
	lines(
		"// i-registry s.r.o. : http://www.i-registry.cz/",
		"// Submitted by Martin Semrad <semrad@i-registry.cz>",
		"co.cz",
	),
	lines(
		"// CDN77.com : http://www.cdn77.com",
		"// Submitted by Jan Krpes <jan.krpes@cdn77.com>",
		"cdn77-storage.com",
		"rsc.contentproxy9.cz",
		"r.cdn77.net",
		"cdn77-ssl.net",
		"c.cdn77.org",
		"rsc.cdn77.org",
		"ssl.origin.cdn77-secure.org",
	),
	lines(
		"// Cloud DNS Ltd : http://www.cloudns.net",
		"// Submitted by Aleksander Hristov <noc@cloudns.net> & Boyan Peychev <boyan@cloudns.net>",
		"cloudns.asia",
		"cloudns.be",
		"cloudns.biz",
		"cloudns.cc",
		"cloudns.ch",
		"cloudns.cl",
		"cloudns.club",
		"dnsabr.com",
		"cloudns.cx",
		"cloudns.eu",
		"cloudns.in",
		"cloudns.info",
		"dns-cloud.net",
		"dns-dynamic.net",
		"cloudns.nz",
		"cloudns.org",
		"cloudns.ph",
		"cloudns.pro",
		"cloudns.pw",
		"cloudns.us",
	),
	lines(
		"// Daplie, Inc : https://daplie.com",
		"// Submitted by AJ ONeal <aj@daplie.com>",
		"daplie.me",
		"localhost.daplie.me",
	),
	lines(
		"// Datto, Inc. : https://www.datto.com/",
		"// Submitted by Philipp Heckel <ph@datto.com>",
		"dattolocal.com",
		"dattorelay.com",
		"dattoweb.com",
		"mydatto.com",
		"dattolocal.net",
		"mydatto.net",
	),
	lines(
		"// Bip : https://bip.sh",
		"// Submitted by Joel Kennedy <joel@bip.sh>",
		"bip.sh",
	),
	lines(
		"// bitbridge.net : Submitted by Craig Welch, abeliidev@gmail.com",
		"bitbridge.net",
	),
	lines(
		"// ddnss.de : https://www.ddnss.de/",
		"// Submitted by Robert Niedziela <webmaster@ddnss.de>",
		"ddnss.de",
		"dyn.ddnss.de",
		"dyndns.ddnss.de",
		"dyn-ip24.de",
		"dyndns1.de",
		"home-webserver.de",
		"dyn.home-webserver.de",
		"myhome-server.de",
		"ddnss.org",
	),
	lines(
		"// Definima : http://www.definima.com/",
		"// Submitted by Maxence Bitterli <maxence@definima.com>",
		"definima.io",
		"definima.net",
	),
	lines(
		"// DigitalOcean App Platform : https://www.digitalocean.com/products/app-platform/",
		"// Submitted by Braxton Huggins <psl-maintainers@digitalocean.com>",
		"ondigitalocean.app",
	),
	lines(
		"// DigitalOcean Spaces : https://www.digitalocean.com/products/spaces/",
		"// Submitted by Robin H. Johnson <psl-maintainers@digitalocean.com>",
		"*.digitaloceanspaces.com",
	),
	lines(
		"// DigitalPlat : https://www.digitalplat.org/",
		"// Submitted by Edward Hsing <contact@digitalplat.org>",
		"us.kg",
	),
	lines(
		"// dnstrace.pro : https://dnstrace.pro/",
		"// Submitted by Chris Partridge <chris@partridge.tech>",
		"bci.dnstrace.pro",
	),
	lines(
		"// ECG Robotics, Inc: https://ecgrobotics.org",
		"// Submitted by <frc1533@ecgrobotics.org>",
		"onred.one",
		"staging.onred.one",
	),
	lines(
		"// Fedora : https://fedoraproject.org/",
		"// submitted by Patrick Uiterwijk <puiterwijk@fedoraproject.org>",
		"fedorainfracloud.org",
		"fedorapeople.org",
		"cloud.fedoraproject.org",
		"app.os.fedoraproject.org",
		"app.os.stg.fedoraproject.org",
	),
	lines(
		"// Frusky MEDIA&PR : https://www.frusky.de",
		"// Submitted by Victor Pupynin <hallo@frusky.de>",
		"*.frusky.de",
	),
	lines(
		"// RavPage : https://www.ravpage.co.il",
		"// Submitted by Roni Horowitz <roni@responder.co.il>",
		"ravpage.co.il",
	),
	lines(
		"// CDDO : https://www.gov.uk/guidance/get-an-api-domain-on-govuk",
		"// Submitted by Jamie Tanna <jamie.tanna@digital.cabinet-office.gov.uk>",
		"api.gov.uk",
	),
	lines(
		"// GOV.UK Platform as a Service : https://www.cloud.service.gov.uk/",
		"// Submitted by Tom Whitwell <gov-uk-paas-support@digital.cabinet-office.gov.uk>",
		"cloudapps.digital",
		"london.cloudapps.digital",
	),
	lines(
		"// GOV.UK Pay : https://www.payments.service.gov.uk/",
		"// Submitted by Richard Baker <richard.baker@digital.cabinet-office.gov.uk>",
		"pymnt.uk",
	),
	lines(
		"// Helio Networks : https://heliohost.org",
		"// Submitted by Ben Frede <admin@heliohost.org>",
		"helioho.st",
		"heliohost.us",
	),
	lines(
		"// Häkkinen.fi",
		"// Submitted by Eero Häkkinen <Eero+psl@Häkkinen.fi>",
		"häkkinen.fi",
	),
	lines(
		"// is-a.dev : https://www.is-a.dev",
		"// Submitted by William Harrison <admin@maintainers.is-a.dev>",
		"is-a.dev",
	),
	lines(
		"// I-O DATA DEVICE, INC. : http://www.iodata.com/",
		"// Submitted by Yuji Minagawa <domains-admin@iodata.jp>",
		"iobb.net",
	),
	lines(
		"// KUROKU LTD : https://kuroku.ltd/",
		"// Submitted by DisposaBoy <security@oya.to>",
		"oya.to",
	),
	lines(
		"// Katholieke Universiteit Leuven: https://www.kuleuven.be",
		"// Submitted by Abuse KU Leuven <abuse@kuleuven.be>",
		"ezproxy.kuleuven.be",
		"kuleuven.cloud",
	),
	lines(
		"// .KRD : http://nic.krd/data/krd/Registration%20Policy.pdf",
		"co.krd",
		"edu.krd",
	),
	lines(
		"// Lokalized : https://lokalized.nl",
		"// Submitted by Noah Taheij <noah@lokalized.nl>",
		"servers.run",
	),
	lines(
		"// May First - People Link : https://mayfirst.org/",
		"// Submitted by Jamie McClelland <info@mayfirst.org>",
		"mayfirst.info",
		"mayfirst.org",
	),
	lines(
		"// mcpe.me : https://mcpe.me",
		"// Submitted by Noa Heyl <hi@noa.dev>",
		"mcpe.me",
	),
	lines(
		"// NFSN, Inc. : https://www.NearlyFreeSpeech.NET/",
		"// Submitted by Jeff Wheelhouse <support@nearlyfreespeech.net>",
		"nfshost.com",
	),
	lines(
		"// NFT.Storage : https://nft.storage/",
		"// Submitted by Vasco Santos <vasco.santos@protocol.ai> or <support@nft.storage>",
		"ipfs.nftstorage.link",
	),
	lines(
		"// No-IP.com : https://noip.com/",
		"// Submitted by Deven Reza <publicsuffixlist@noip.com>",
		"mmafan.biz",
		"myftp.biz",
		"no-ip.biz",
		"no-ip.ca",
		"fantasyleague.cc",
		"gotdns.ch",
		"3utilities.com",
		"blogsyte.com",
		"ciscofreak.com",
		"damnserver.com",
		"ddnsking.com",
		"ditchyourip.com",
		"dnsiskinky.com",
		"dynns.com",
		"geekgalaxy.com",
		"health-carereform.com",
		"homesecuritymac.com",
		"homesecuritypc.com",
		"myactivedirectory.com",
		"mysecuritycamera.com",
		"myvnc.com",
		"net-freaks.com",
		"onthewifi.com",
		"point2this.com",
		"quicksytes.com",
		"securitytactics.com",
		"servebeer.com",
		"servecounterstrike.com",
		"serveexchange.com",
		"serveftp.com",
		"servegame.com",
		"servehalflife.com",
		"servehttp.com",
		"servehumour.com",
		"serveirc.com",
		"servemp3.com",
		"servep2p.com",
		"servepics.com",
		"servequake.com",
		"servesarcasm.com",
		"stufftoread.com",
		"unusualperson.com",
		"workisboring.com",
		"dvrcam.info",
		"ilovecollege.info",
		"no-ip.info",
		"brasilia.me",
		"ddns.me",
		"dnsfor.me",
		"hopto.me",
		"loginto.me",
		"noip.me",
		"webhop.me",
		"bounceme.net",
		"ddns.net",
		"eating-organic.net",
		"mydissent.net",
		"myeffect.net",
		"mymediapc.net",
		"mypsx.net",
		"mysecuritycamera.net",
		"nhlfan.net",
		"no-ip.net",
		"pgafan.net",
		"privatizehealthinsurance.net",
		"redirectme.net",
		"serveblog.net",
		"serveminecraft.net",
		"sytes.net",
		"cable-modem.org",
		"collegefan.org",
		"couchpotatofries.org",
		"hopto.org",
		"mlbfan.org",
		"myftp.org",
		"mysecuritycamera.org",
		"nflfan.org",
		"no-ip.org",
		"read-books.org",
		"ufcfan.org",
		"zapto.org",
		"no-ip.co.uk",
		"golffan.us",
		"noip.us",
		"pointto.us",
	),
	lines(
		"// NodeArt : https://nodeart.io",
		"// Submitted by Konstantin Nosov <Nosov@nodeart.io>",
		"stage.nodeart.io",
	),
	lines(
		"// One.com: https://www.one.com/",
		"// Submitted by Jacob Bunk Nielsen <jbn@one.com>",
		"123webseite.at",
		"123website.be",
		"simplesite.com.br",
		"123website.ch",
		"simplesite.com",
		"123webseite.de",
		"123hjemmeside.dk",
		"123miweb.es",
		"123kotisivu.fi",
		"123siteweb.fr",
		"simplesite.gr",
		"123homepage.it",
		"123website.lu",
		"123website.nl",
		"123hjemmeside.no",
		"service.one",
		"simplesite.pl",
		"123paginaweb.pt",
		"123minsida.se",
	),
	lines(
		"// .pl domains (grandfathered)",
		"art.pl",
		"gliwice.pl",
		"krakow.pl",
		"poznan.pl",
		"wroc.pl",
		"zakopane.pl",
	),
	lines(
		"// Pantheon Systems, Inc. : https://pantheon.io/",
		"// Submitted by Gary Dylina <gary@pantheon.io>",
		"gotpantheon.com",
		"pantheonsite.io",
	),
	lines(
		"// PE Ulyanov Kirill Sergeevich : https://airy.host",
		"// Submitted by Kirill Ulyanov <k.ulyanov@airy.host>",
		"lk3.ru",
	),
	lines(
		"// Rad Web Hosting: https://radwebhosting.com",
		"// Submitted by Scott Claeys <s.claeys@radwebhosting.com>",
		"cloudsite.builders",
		"myradweb.net",
		"servername.us",
	),
	lines(
		"// Raidboxes GmbH : https://raidboxes.de",
		"// Submitted by Auke Tembrink <hostmaster@raidboxes.de>",
		"myrdbx.io",
		"site.rb-hosting.io",
	),
	lines(
		"// Redgate Software: https://red-gate.com",
		"// Submitted by Andrew Farries <andrew.farries@red-gate.com>",
		"instances.spawn.cc",
	),
	lines(
		"// Redstar Consultants : https://www.redstarconsultants.com/",
		"// Submitted by Jons Slemmer <jons@redstarconsultants.com>",
		"instantcloud.cn",
	),
	lines(
		"// Russian Academy of Sciences",
		"// Submitted by Tech Support <support@rasnet.ru>",
		"ras.ru",
	),
	lines(
		"// QA2",
		"// Submitted by Daniel Dent (https://www.danieldent.com/)",
		"qa2.com",
	),
	lines(
		"// QCX",
		"// Submitted by Cassandra Beelen <cassandra@beelen.one>",
		"qcx.io",
		"*.sys.qcx.io",
	),
	lines(
		"// QNAP System Inc : https://www.qnap.com",
		"// Submitted by Nick Chang <cloudadmin@qnap.com>",
		"myqnapcloud.cn",
		"alpha-myqnapcloud.com",
		"dev-myqnapcloud.com",
		"mycloudnas.com",
		"mynascloud.com",
		"myqnapcloud.com",
	),
	lines(
		"// Senseering GmbH : https://www.senseering.de",
		"// Submitted by Felix Mönckemeyer <f.moenckemeyer@senseering.de>",
		"senseering.net",
	),
	lines(
		"// Smallregistry by Promopixel SARL: https://www.smallregistry.net",
		"// Former AFNIC's SLDs",
		"// Submitted by Jérôme Lipowicz <support@promopixel.com>",
		"aeroport.fr",
		"avocat.fr",
		"chambagri.fr",
		"chirurgiens-dentistes.fr",
		"experts-comptables.fr",
		"medecin.fr",
		"notaires.fr",
		"pharmacien.fr",
		"port.fr",
		"veterinaire.fr",
	),
	lines(
		"// staticland : https://static.land",
		"// Submitted by Seth Vincent <sethvincent@gmail.com>",
		"static.land",
		"dev.static.land",
		"sites.static.land",
	),
	lines(
		"// Storebase : https://www.storebase.io",
		"// Submitted by Tony Schirmer <tony@storebase.io>",
		"storebase.store",
	),
	lines(
		"// Strapi : https://strapi.io/",
		"// Submitted by Florent Baldino <security@strapi.io>",
		"strapiapp.com",
		"media.strapiapp.com",
	),
	lines(
		"// Strategic System Consulting (eApps Hosting): https://www.eapps.com/",
		"// Submitted by Alex Oancea <aoancea@cloudscale365.com>",
		"vps-host.net",
		"atl.jelastic.vps-host.net",
		"njs.jelastic.vps-host.net",
		"ric.jelastic.vps-host.net",
	),
	lines(
		"// Sony Interactive Entertainment LLC : https://sie.com/",
		"// Submitted by David Coles <david.coles@sony.com>",
		"playstation-cloud.com",
	),
	lines(
		"// SourceLair PC : https://www.sourcelair.com",
		"// Submitted by Antonis Kalipetis <akalipetis@sourcelair.com>",
		"apps.lair.io",
		"*.stolos.io",
	),
	lines(
		"// SpaceKit : https://www.spacekit.io/",
		"// Submitted by Reza Akhavan <spacekit.io@gmail.com>",
		"spacekit.io",
	),
	lines(
		"// SpeedPartner GmbH: https://www.speedpartner.de/",
		"// Submitted by Stefan Neufeind <info@speedpartner.de>",
		"customer.speedpartner.de",
	),
	lines(
		"// Spreadshop (sprd.net AG) : https://www.spreadshop.com/",
		"// Submitted by Martin Breest <security@spreadshop.com>",
		"myspreadshop.at",
		"myspreadshop.com.au",
		"myspreadshop.be",
		"myspreadshop.ca",
		"myspreadshop.ch",
		"myspreadshop.com",
		"myspreadshop.de",
		"myspreadshop.dk",
		"myspreadshop.es",
		"myspreadshop.fi",
		"myspreadshop.fr",
		"myspreadshop.ie",
		"myspreadshop.it",
		"myspreadshop.net",
		"myspreadshop.nl",
		"myspreadshop.no",
		"myspreadshop.pl",
		"myspreadshop.se",
		"myspreadshop.co.uk",
	),
	lines(
		"// Studenten Net Twente : http://www.snt.utwente.nl/",
		"// Submitted by Silke Hofstra <syscom@snt.utwente.nl>",
		"utwente.io",
	),
	lines(
		"// UNIVERSAL DOMAIN REGISTRY : https://www.udr.org.yt/",
		"// see also: whois -h whois.udr.org.yt help",
		"// Submitted by Atanunu Igbunuroghene <publicsuffixlist@udr.org.yt>",
		"name.pm",
		"sch.tf",
		"biz.wf",
		"sch.wf",
		"org.yt",
	),
	lines(
		"// .US",
		"// Submitted by Ed Moore <Ed.Moore@lib.de.us>",
		"lib.de.us",
	),
	lines(
		"// VeryPositive SIA : http://very.lv",
		"// Submitted by Danko Aleksejevs <danko@very.lv>",
		"2038.io",
	),
	lines(
		"// V.UA Domain Administrator : https://domain.v.ua/",
		"// Submitted by Serhii Rostilo <sergey@rostilo.kiev.ua>",
		"v.ua",
	),
}
